//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <crtdbg.h>
#include <math.h>

#include "VBitmap.h"
#include "Error.h"

#include "SceneDetector.h"

SceneDetector::SceneDetector(PixDim width, PixDim height) {
	cur_lummap = last_lummap = NULL;
	last_valid = FALSE;
	first_diff = TRUE;

	cut_threshold = 50 * tile_w * tile_h;
	fade_threshold = 4 * tile_w * tile_h;
	use_alternate = FALSE;
	multiplier = 35;

	try {
		tile_w = (width + 7)/8;
		tile_h = (height + 7)/8;

		if (	!(cur_lummap  = new Pixel[tile_h * tile_w])
			||	!(last_lummap = new Pixel[tile_h * tile_w]))

			throw MyMemoryError();
	} catch(...) {
		_destruct();
		throw;
	}

	worst_block = 99.0;
	last_psnr = 99.0;

}

SceneDetector::SceneDetector(PixDim width, PixDim height, bool bHighDensity) {
// actually, bHighDensity is ignored - NN

	cur_lummap = last_lummap = NULL;
	last_valid = FALSE;
	first_diff = TRUE;

	cut_threshold = 50 * tile_w * tile_h;
	fade_threshold = 4 * tile_w * tile_h;
	use_alternate = FALSE;
	multiplier = 35;

	try {
		tile_w = (width + 3)/4;
		tile_h = (height + 3)/4;

		if (	!(cur_lummap  = new Pixel[tile_h * tile_w])
			||	!(last_lummap = new Pixel[tile_h * tile_w]))

			throw MyMemoryError();
	} catch(...) {
		_destruct();
		throw;
	}
}

SceneDetector::_destruct() {
	delete[] cur_lummap;
	delete[] last_lummap;
}

SceneDetector::~SceneDetector() {
	_destruct();
}

//////////////////////////////////////////////////////////////////////////

void SceneDetector::SetThresholds(int cut_threshold, int fade_threshold, bool bAlt, int multiplier) {
	this->cut_threshold		= (cut_threshold * tile_w * tile_h)/16;
	this->fade_threshold	= (fade_threshold * tile_w * tile_h * 3)/16.0;
	this->use_alternate		= bAlt;
	this->multiplier		= multiplier;
}

BOOL SceneDetector::Submit(VBitmap *vbm) {

	if( use_alternate ) return Submit2(vbm);

	long last_frame_diffs = 0;
	long lum_total = 0;
	double lum_sq_total = 0.0;
	long len = tile_w * tile_h;

	if (vbm->w > tile_w*8 || vbm->h > tile_h*8 || !cut_threshold || !fade_threshold)
		return FALSE;

	FlipBuffers();
	BitmapToLummap(cur_lummap, vbm);

	if (!last_valid) {
		last_valid = TRUE;
		return FALSE;
	}

/////////////

	Pixel *t1 = cur_lummap, *t2 = last_lummap;
	BOOL is_fade;

	do {
		Pixel c1 = *t1++;
		Pixel c2 = *t2++;

		last_frame_diffs +=	  abs((int)(c2>>16)-(int)(c1>>16))
							+ abs((int)((c2>>8)&255) -(int)((c1>>8)&255))
							+ abs((int)(c2&255)-(int)(c1&255));

		long lum;
		if( vbm->isYUY2 ) {
			lum = (c1>>8)&255;
		} else {
			lum = ((c1>>16)*54 + ((c1>>8)&255)*183 + (c1&255)*19 + 128)>>8;
		}

		lum_total += lum;
		lum_sq_total += (double)lum * (double)lum;
	} while(--len);

	lum_sq_total *= tile_w*tile_h;
	// seems that adding diffs between YUY2 images produces far less amplitude...
	if( vbm->isYUY2 ) last_frame_diffs *= 4;


//	_RPT3(0,"Last frame diffs=%ld, lum(linear)=%ld, lum(rms)=%f\n",last_frame_diffs,lum_total,sqrt(lum_sq_total));

	if (fade_threshold) {
		is_fade = fabs(sqrt(lum_sq_total) - (double)lum_total) < fade_threshold;

		if (first_diff) {
			last_fade_state = is_fade;
		} else {
			// If we've encountered a new fade, return 'scene changed'

			if (!last_fade_state && is_fade) return TRUE;

			// Hit the end of an initial fade?

			if (last_fade_state && !is_fade)
				last_fade_state = FALSE;
		}
	}

	// Cut/dissolve detection

	return cut_threshold ? last_frame_diffs > cut_threshold : FALSE;
}


BOOL SceneDetector::Submit2(VBitmap *vbm) {

// Easy 'variable threshold'/'motion based' scenedetection 
// added by ViRA
//
//   Keyframe = AbsDiff > (Prev_AbsDiff * Multiplier) ? true : false;
//
// Notes:
//   - Most fades are already catched, fade detection taken out
//	 - Disregards any threshold settings used in virtualdub
//   - Quick hack, only tested in mpg2avi
// 
//======================================
// Room for improvements:
//		- mmx optimize absolute difference
//		- variable difference multiplier 
//		  (*4 misses some lowmotion scenechanges,
//         *3 misspredicts scenes at highmotion scenes,
//	       *3 works better for cartoons)
//======================================
// 
// Branchless MMX optimized Abs() on 8xY at a time:
//
//	movq       mm0, DWORD PTR [lummap1]		// grab 8 Y pixels of this frame
//	movq       mm1, DWORD PTR [lummap2]		// grab 8 Y pixels of previous frame
//	movq       mm2, mm0					    // make a copy for abs diff operation	
//	psubusb    mm0, mm1						// do subtraction one way (w/saturation)
//	psubusb    mm1, mm2						// do subtraction the other way (w/saturation)	
//	por        mm1, mm0						// merge results of 1st half
//
	
	static long last_frame_diffs = 0; 
	long frame_diffs = 0;
	bool ret;

	long len = tile_w * tile_h;

	if (vbm->w > tile_w*8 || vbm->h > tile_h*8)
		return FALSE;

	FlipBuffers();
	BitmapToLummap(cur_lummap, vbm);

	if (!last_valid) {
		last_valid = TRUE;
		return FALSE;
	}

/////////////

	Pixel *t1 = cur_lummap, *t2 = last_lummap;
	
	do {
		Pixel c1 = *t1++;
		Pixel c2 = *t2++;

		frame_diffs +=	abs((int)(c2>>16)-(int)(c1>>16))
							+ abs((int)((c2>>8)&255) -(int)((c1>>8)&255))
							+ abs((int)(c2&255)-(int)(c1&255));
	} while(--len);

	// seems that adding diffs between YUY2 images produces far less amplitude...
	if( vbm->isYUY2 ) frame_diffs *= 4;

	ret = frame_diffs > last_frame_diffs ? true : false;
	last_frame_diffs = MulDiv( frame_diffs, multiplier, 10 );
	
	//added for 'stills' not to be coded as keyframes
	//but also makes end-of-fades uncatchable
	//worth dropping these frames?
	if (last_frame_diffs < 10000) last_frame_diffs = 10000;

	return ret;

}






// Let's see if we can use this object for another purpose :
// -> detect DivX encoding errors
//
// VideoSequenceCompressor will instantiate and use it this way
// - Reset()
// - Submit() the uncompressed bitmap
// - CheckForErrors() the decompressed bitmap


void SceneDetector::SetQThreshold(int q_threshold) {
	this->q_threshold		= q_threshold;
}


//
// 06/12/2001 : 
//		function takes now 2 parameters, the old and the new bitmap.
//		resolution of the pre-posterization is now 4x4 (was 8x8)
//		comparison algorythm is the real PSNR.
//

BOOL SceneDetector::CheckForErrors(VBitmap *ovbm, VBitmap *vbm) {
	long lum_diff, tot_diff;
	long max_diff = 0;
	long len = tile_w * tile_h;

	if (ovbm->w > tile_w*4 || ovbm->h > tile_h*4)
		return FALSE;

	FlipBuffers();
	BitmapToLummap4x4(cur_lummap, ovbm);
	last_valid = TRUE;

	if (vbm->w > tile_w*4 || vbm->h > tile_h*4)
		return FALSE;

	FlipBuffers();
	BitmapToLummap4x4(cur_lummap, vbm);

	if (!last_valid) {
		last_valid = TRUE;
		return FALSE;
	}

/////////////

	Pixel *t1 = cur_lummap, *t2 = last_lummap;

	tot_diff = 0;
	do {
		Pixel c1 = *t1++;
		Pixel c2 = *t2++;

// All this doesn't seem satisfactory...
/*
		lum_diff =	  abs((int)(c2>>16)-(int)(c1>>16))
					+ abs((int)((c2>>8)&255) -(int)((c1>>8)&255))
					+ abs((int)(c2&255)-(int)(c1&255));
*/
/*
		lum_diff = abs( (((c1>>16)*54 + ((c1>>8)&255)*183 + (c1&255)*19 + 128)>>8)
			           -(((c2>>16)*54 + ((c2>>8)&255)*183 + (c2&255)*19 + 128)>>8) );

		if( lum_diff>max_diff ) max_diff=lum_diff;
*/
// how about PSNR ?
		lum_diff =	  ((int)(c2>>16)-(int)(c1>>16))*((int)(c2>>16)-(int)(c1>>16))
					+ ((int)((c2>>8)&255)-(int)((c1>>8)&255))*((int)((c2>>8)&255)-(int)((c1>>8)&255))
					+ ((int)(c2&255)-(int)(c1&255))*((int)(c2&255)-(int)(c1&255));

		if( lum_diff>max_diff ) max_diff=lum_diff;
		tot_diff += lum_diff;

	} while(--len);

	if( max_diff==0 ) {
		max_diff = 1;
		tot_diff = 1;
	}

	double y = .5+10*log((195075.0*tile_w*tile_h)/(tot_diff*1.0))/log(10); // 195075.0==255*255*3
	double x = .5+10*log((255*255*3)/(max_diff*1.0))/log(10);

	worst_block = x;
	last_psnr = y;

	return worst_block<q_threshold;

/*
	last_max_diff = lum_diff;

	return max_diff>q_threshold;
*/
}



double SceneDetector::LumaNoise(VBitmap *vbm) {
	long lum_total = 0;
	double lum_sq_total = 0.0;
	long len = tile_w * tile_h;

	if (vbm->w > tile_w*8 || vbm->h > tile_h*8)
		return FALSE;

	FlipBuffers();
	BitmapToLummap(cur_lummap, vbm);

/////////////

	Pixel *t1 = cur_lummap;

	do {
		Pixel c1 = *t1++;

//		long lum = ((c1>>16)*54 + ((c1>>8)&255)*183 + (c1&255)*19 + 128)>>8;
		long lum;
		if( vbm->isYUY2 )
			lum = (c1>>8)&255;
		else
			lum = ((c1>>16)*54 + ((c1>>8)&255)*183 + (c1&255)*19 + 128)>>8; // dividee

		lum_total += lum;
		lum_sq_total += (double)lum * (double)lum;
	} while(--len);

	lum_sq_total *= tile_w*tile_h;

	return fabs( sqrt(lum_sq_total) - (double)lum_total ) / (tile_w*tile_h);

}


void SceneDetector::Reset() {
	last_valid = FALSE;
}

//////////////////////////////////////////////////////////////////////////
//
// Note: these names should be changed, since these functions do NOT relate
//	     to LUMinance at all. There are just fast 8x8 block averaging.
//

extern "C" Pixel __cdecl asm_scene_lumtile32(void *src, long w, long h, long pitch);
extern "C" Pixel __cdecl asm_scene_lumtile24(void *src, long w, long h, long pitch);
extern "C" Pixel __cdecl asm_scene_lumtile16(void *src, long w, long h, long pitch);

extern "C" Pixel __cdecl asm_scene_lumtile_yuy2(void *src, long w, long h, long pitch);

void SceneDetector::BitmapToLummap(Pixel *lummap, VBitmap *vbm) {
	char *src, *src_row = (char *)vbm->data;
	int mh = 8;
	long w,h;

	h = (vbm->h+7)/8;

	switch(vbm->depth) {
	case 32:
		do {
			if (h<=1 && (vbm->h&7)) mh = vbm->h&7;

			src = src_row; src_row += vbm->pitch*8;
			w = vbm->w/8;
			do {
				*lummap++ = asm_scene_lumtile32(src, 8, mh, vbm->pitch);
				src += 32;
			} while(--w);

			if (vbm->w & 7) {
				*lummap++ = asm_scene_lumtile32(src, vbm->w&7, mh, vbm->pitch);
			}
		} while(--h);
		break;
	case 24:
		do {
			if (h<=1 && (vbm->h&7)) mh = vbm->h&7;

			src = src_row; src_row += vbm->pitch*8;
			w = vbm->w/8;
			do {
				*lummap++ = asm_scene_lumtile24(src, 8, mh, vbm->pitch);
				src += 24;
			} while(--w);

			if (vbm->w & 7) {
				*lummap++ = asm_scene_lumtile24(src, vbm->w&7, mh, vbm->pitch);
			}
		} while(--h);
		break;
	case 16:
		if( vbm->isYUY2 ) {
			do {
				if (h<=1 && (vbm->h&7)) mh = vbm->h&7;

				src = src_row; src_row += vbm->pitch*8;
				w = vbm->w/8;
				do {
					*lummap++ = asm_scene_lumtile_yuy2(src, 4, mh, vbm->pitch);
					src += 16;
				} while(--w);

				if (vbm->w & 6) {
					*lummap++ = asm_scene_lumtile_yuy2(src, (vbm->w&6)/2, mh, vbm->pitch);
				}
			} while(--h);
		} else {
			do {
				if (h<=1 && (vbm->h&7)) mh = vbm->h&7;

				src = src_row; src_row += vbm->pitch*8;
				w = vbm->w/8;
				do {
					*lummap++ = asm_scene_lumtile16(src, 4, mh, vbm->pitch);
					src += 16;
				} while(--w);

				if (vbm->w & 6) {
					*lummap++ = asm_scene_lumtile16(src, (vbm->w&6)/2, mh, vbm->pitch);
				}
			} while(--h);
		}
		break;
	}
}

//
// the 4x4 clones
// (note: depth=16 untested)
//

extern "C" Pixel __cdecl asm_scene_lumtile32_4x4(void *src, long w, long h, long pitch);
extern "C" Pixel __cdecl asm_scene_lumtile24_4x4(void *src, long w, long h, long pitch);
extern "C" Pixel __cdecl asm_scene_lumtile16_4x4(void *src, long w, long h, long pitch);

void SceneDetector::BitmapToLummap4x4(Pixel *lummap, VBitmap *vbm) {
	char *src, *src_row = (char *)vbm->data;
	int mh = 4;
	long w,h;

	h = (vbm->h+3)/4;

	switch(vbm->depth) {
	case 32:
		do {
			if (h<=1 && (vbm->h&3)) mh = vbm->h&3;

			src = src_row; src_row += vbm->pitch*4;
			w = vbm->w/4;
			do {
				*lummap++ = asm_scene_lumtile32_4x4(src, 4, mh, vbm->pitch);
				src += 16;
			} while(--w);

			if (vbm->w & 3) {
				*lummap++ = asm_scene_lumtile32_4x4(src, vbm->w&3, mh, vbm->pitch);
			}
		} while(--h);
		break;
	case 24:
		do {
			if (h<=1 && (vbm->h&3)) mh = vbm->h&3;

			src = src_row; src_row += vbm->pitch*4;
			w = vbm->w/4;
			do {
				*lummap++ = asm_scene_lumtile24_4x4(src, 4, mh, vbm->pitch);
				src += 12;
			} while(--w);

			if (vbm->w & 3) {
				*lummap++ = asm_scene_lumtile24_4x4(src, vbm->w&3, mh, vbm->pitch);
			}
		} while(--h);
		break;
	case 16:
		do {
			if (h<=1 && (vbm->h&3)) mh = vbm->h&3;

			src = src_row; src_row += vbm->pitch*4;
			w = vbm->w/4;
			do {
				*lummap++ = asm_scene_lumtile16_4x4(src, 2, mh, vbm->pitch);
				src += 8;
			} while(--w);

			if (vbm->w & 2) {
				*lummap++ = asm_scene_lumtile16_4x4(src, (vbm->w&2)/2, mh, vbm->pitch);
			}
		} while(--h);
		break;
	}
}

void SceneDetector::FlipBuffers() {
	Pixel *t;

	t = cur_lummap;
	cur_lummap = last_lummap;
	last_lummap = t;
}
